import json
import asyncio
import logging
from fastmcp import Client, FastMCP

from frankfurtermcp.server import FrankfurterMCP
import pytest

logger = logging.getLogger(__name__)


class TestMCPServer:
    @classmethod
    @pytest.fixture(scope="class", autouse=True)
    def mcp_server(cls):
        """
        Fixture to register features in an MCP server.
        """
        server = FastMCP()
        mcp_obj = FrankfurterMCP()
        server_with_features = mcp_obj.register_features(server)
        return server_with_features

    @classmethod
    @pytest.fixture(scope="class", autouse=True)
    def mcp_client(cls, mcp_server):
        """
        Fixture to create a client for the MCP server.
        """
        mcp_client = Client(
            transport=mcp_server,
            timeout=60,
        )
        return mcp_client

    async def call_tool(self, tool_name: str, mcp_client: Client, **kwargs):
        """
        Helper method to call a tool on the MCP server.
        """
        logger.debug(f"Calling tool '{tool_name}' with arguments: {kwargs}")
        async with mcp_client:
            result = await mcp_client.call_tool(tool_name, arguments=kwargs)
            await mcp_client.close()
        logger.debug(f"Tool '{tool_name}' returned result: {result}")
        return result

    def test_get_supported_currencies(self, mcp_client):
        """
        Test the get_supported_currencies function to ensure it returns a list of supported currencies.
        """
        test_method = "get_supported_currencies"
        response = asyncio.run(
            self.call_tool(
                tool_name=test_method,
                mcp_client=mcp_client,
            )
        )
        json_result: dict = json.loads(response.content[0].text)
        assert len(json_result.keys()) > 0, "Expected non-empty list of currencies"
        assert all(
            (isinstance(code, str) and len(code) == 3) for code in json_result.keys()
        ), "All currency codes should be 3-character strings"

    def test_convert_currency_latest(self, mcp_client):
        """
        Test the convert_currency_latest function to ensure it returns a list of supported currencies.
        """
        test_method = "convert_currency_latest"
        response = asyncio.run(
            self.call_tool(
                tool_name=test_method,
                mcp_client=mcp_client,
                from_currency="GBP",
                to_currency="JPY",
                amount=100.0,
            )
        )
        json_result: dict = json.loads(response.content[0].text)
        logger.info(f"{test_method} response: {json_result}")
        assert isinstance(json_result["converted_amount"], float), (
            "Expected float value for converted amount"
        )
        assert json_result["converted_amount"] > 100.0, (
            "The exchange rate for GBP to JPY should be greater than 1.0"
        )

    def test_get_latest_exchange_rates(self, mcp_client):
        """
        Test the get_latest_exchange_rates function to ensure that it returns the list of latest rates with other currencies.
        """
        test_method = "get_latest_exchange_rates"
        response = asyncio.run(
            self.call_tool(
                tool_name=test_method,
                mcp_client=mcp_client,
                base_currency="JPY",
                symbols=["EUR", "GBP", "CHF", "NZD"],
            )
        )
        json_result: dict = json.loads(response.content[0].text)
        assert len(json_result["rates"].keys()) > 0, (
            "Expected non-empty list of currency rates"
        )
        assert all(
            (isinstance(code, str) and len(code) == 3)
            for code in json_result["rates"].keys()
        ), "All currency codes for exchange rates should be 3-character strings"

    def test_get_historical_exchange_rates(self, mcp_client):
        """
        Test the get_historical_exchange_rates function to ensure that it returns the list of historical rates with other currencies.
        """
        test_method = "get_historical_exchange_rates"
        response = asyncio.run(
            self.call_tool(
                tool_name=test_method,
                mcp_client=mcp_client,
                base_currency="JPY",
                start_date="2025-06-01",
                end_date="2025-06-19",
                symbols=["EUR", "GBP", "CHF", "NZD"],
            )
        )
        json_result: dict = json.loads(response.content[0].text)
        assert all(
            len(rates_for_date) > 0
            for _, rates_for_date in json_result["rates"].items()
        ), "Expected non-empty list of currency rates"
        assert all(
            (
                (isinstance(code, str) and len(code) == 3)
                for code in rates_for_date.keys()
            )
            for _, rates_for_date in json_result["rates"].items()
        ), "All currency codes for exchange rates should be 3-character strings"

    def test_get_latest_exchange_rates_for_single_currency(self, mcp_client):
        """
        Test the get_latest_exchange_rates function to ensure that it returns the latest rates for a single currency.
        """
        test_method = "get_latest_exchange_rates"
        response = asyncio.run(
            self.call_tool(
                tool_name=test_method,
                mcp_client=mcp_client,
                base_currency="JPY",
                symbols="GBP",
            )
        )
        json_result: dict = json.loads(response.content[0].text)
        assert len(json_result["rates"].keys()) > 0, (
            "Expected non-empty list of currency rates"
        )
        assert all(
            (isinstance(code, str) and len(code) == 3)
            for code in json_result["rates"].keys()
        ), "All currency codes for exchange rates should be 3-character strings"

    def test_get_historical_exchange_rates_for_a_single_currency(self, mcp_client):
        """
        Test the get_historical_exchange_rates function to ensure that it returns the historical rates for a single currency.
        """
        test_method = "get_historical_exchange_rates"
        response = asyncio.run(
            self.call_tool(
                tool_name=test_method,
                mcp_client=mcp_client,
                base_currency="JPY",
                start_date="2025-06-01",
                end_date="2025-06-19",
                symbols=["EUR", "GBP", "CHF", "NZD"],
            )
        )
        json_result: dict = json.loads(response.content[0].text)
        assert all(
            len(rates_for_date) > 0
            for _, rates_for_date in json_result["rates"].items()
        ), "Expected non-empty list of currency rates"
        assert all(
            (
                (isinstance(code, str) and len(code) == 3)
                for code in rates_for_date.keys()
            )
            for _, rates_for_date in json_result["rates"].items()
        ), "All currency codes for exchange rates should be 3-character strings"

    def test_convert_currency_specific_date(self, mcp_client):
        """
        Test the convert_currency_specific_date function to ensure it returns a list of supported currencies.
        """
        test_method = "convert_currency_specific_date"
        response = asyncio.run(
            self.call_tool(
                tool_name=test_method,
                mcp_client=mcp_client,
                from_currency="GBP",
                to_currency="JPY",
                amount=100.0,
                specific_date="2025-06-01",
            )
        )
        json_result: dict = json.loads(response.content[0].text)
        print(f"{test_method} response: {json_result}")
        assert isinstance(json_result["converted_amount"], float), (
            "Expected float value for converted amount"
        )
        assert json_result["converted_amount"] > 100.0, (
            "The exchange rate for GBP to JPY should be greater than 1.0"
        )



import logging
from environs import Env
from rich.logging import RichHandler

from frankfurtermcp.common import EnvVar


try:
    from icecream import ic

    ic.configureOutput(includeContext=True)
except ImportError:  # pragma: no cover
    # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


env = Env()
env.read_env()

logging.basicConfig(
    level=env.str(EnvVar.LOG_LEVEL, default=EnvVar.DEFAULT__LOG_LEVEL).upper(),
    format="%(message)s",
    datefmt="[%X]",
    handlers=[
        RichHandler(
            rich_tracebacks=False, markup=True, show_path=False, show_time=False
        )
    ],
)



from importlib.metadata import metadata


class EnvVar:
    """
    List of environment variables used in this project.
    """

    LOG_LEVEL = "LOG_LEVEL"
    DEFAULT__LOG_LEVEL = "INFO"

    FASTMCP_HOST = "FASTMCP_HOST"
    DEFAULT__FASTMCP_HOST = "localhost"

    FASTMCP_PORT = "FASTMCP_PORT"
    DEFAULT__FASTMCP_PORT = 8000

    MCP_SERVER_TRANSPORT = "MCP_SERVER_TRANSPORT"
    DEFAULT__MCP_SERVER_TRANSPORT = "stdio"
    ALLOWED__MCP_SERVER_TRANSPORT = ["stdio", "sse", "streamable-http"]

    MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE = "MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE"
    DEFAULT__MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE = True

    FRANKFURTER_API_URL = "FRANKFURTER_API_URL"
    DEFAULT__FRANKFURTER_API_URL = "https://api.frankfurter.dev/v1"

    HTTPX_TIMEOUT = "HTTPX_TIMEOUT"
    DEFAULT__HTTPX_TIMEOUT = 5.0

    HTTPX_VERIFY_SSL = "HTTPX_VERIFY_SSL"
    DEFAULT__HTTPX_VERIFY_SSL = True


class AppMetadata:
    """
    Metadata for the application.
    """

    PACKAGE_NAME = "frankfurtermcp"
    TEXT_CONTENT_META_PREFIX = f"{PACKAGE_NAME}."
    package_metadata = metadata(PACKAGE_NAME)



from datetime import date
import logging
import sys
from typing import Annotated, List, Optional, Union
import httpx

from fastmcp import FastMCP, Context

from frankfurtermcp import env
from marshmallow.validate import OneOf

from pydantic import Field, PositiveFloat
from pydantic_extra_types.currency_code import ISO4217
from frankfurtermcp.common import EnvVar

from frankfurtermcp.common import AppMetadata
from frankfurtermcp.mixin import HTTPHelperMixin, MCPMixin
from frankfurtermcp.model import CurrencyConversionResponse

logger = logging.getLogger(__name__)


class FrankfurterMCP(MCPMixin, HTTPHelperMixin):
    """
    A FastMCP application that provides currency exchange rate functionalities
    using the Frankfurter API.
    """

    tools = [
        {
            "fn": "get_supported_currencies",
            "tags": ["currency-rates", "supported-currencies"],
            "annotations": {
                "readOnlyHint": True,
                "openWorldHint": True,
            },
        },
        {
            "fn": "get_latest_exchange_rates",
            "tags": ["currency-rates", "exchange-rates"],
            "annotations": {
                "readOnlyHint": True,
                "openWorldHint": True,
            },
        },
        {
            "fn": "convert_currency_latest",
            "tags": ["currency-rates", "currency-conversion"],
            "annotations": {
                "readOnlyHint": True,
                "openWorldHint": True,
            },
        },
        {
            "fn": "get_historical_exchange_rates",
            "tags": ["currency-rates", "historical-exchange-rates"],
            "annotations": {
                "readOnlyHint": True,
                "openWorldHint": True,
            },
        },
        {
            "fn": "convert_currency_specific_date",
            "tags": [
                "currency-rates",
                "currency-conversion",
                "historical-exchange-rates",
            ],
            "annotations": {
                "readOnlyHint": True,
                "openWorldHint": True,
            },
        },
    ]

    async def get_supported_currencies(self, ctx: Context):
        """
        Returns a list of three-letter currency codes for the supported currencies.
        """
        try:
            with self.get_httpx_client() as client:
                await ctx.info(
                    f"Fetching supported currencies from Frankfurter API at {self.frankfurter_api_url}"
                )
                http_response = client.get(f"{self.frankfurter_api_url}/currencies")
                http_response.raise_for_status()
                result = http_response.json()
                return self.get_response_text_content(
                    response=result, http_response=http_response
                )
        except httpx.RequestError as e:
            raise ValueError(
                f"Failed to fetch supported currencies from {self.frankfurter_api_url}. {e}"
            )

    def _get_latest_exchange_rates(
        self,
        base_currency: Union[str, None] = None,
        symbols: Union[List[str], None] = None,
    ):
        """
        Internal function to get the latest exchange rates.
        This is a helper function for the main tool.
        """
        try:
            params = {}
            if base_currency:
                params["base"] = base_currency
            if symbols:
                params["symbols"] = ",".join(symbols)
            with self.get_httpx_client() as client:
                http_response = client.get(
                    f"{self.frankfurter_api_url}/latest",
                    params=params,
                )
                http_response.raise_for_status()
                result = http_response.json()
                return result, http_response
        except httpx.RequestError as e:
            raise ValueError(
                f"Failed to fetch latest exchange rates from {self.frankfurter_api_url}. {e}"
            )

    def _get_historical_exchange_rates(
        self,
        specific_date: Union[str, None] = None,
        start_date: Union[str, None] = None,
        end_date: Union[str, None] = None,
        base_currency: Union[str, None] = None,
        symbols: Union[List[str], None] = None,
    ):
        """
        Internal function to get historical exchange rates.
        This is a helper function for the main tool.
        """
        try:
            params = {}
            if base_currency:
                params["base"] = base_currency
            if symbols:
                params["symbols"] = ",".join(symbols)

            frankfurter_url = self.frankfurter_api_url
            if start_date and end_date:
                frankfurter_url += f"/{start_date}..{end_date}"
            elif start_date:
                # If only start_date is provided, we assume the end date is the latest available date
                frankfurter_url += f"/{start_date}.."
            elif specific_date:
                # If only specific_date is provided, we assume it is the date for which we want the rates
                frankfurter_url += f"/{specific_date}"
            else:
                raise ValueError(
                    "You must provide either a specific date, a start date, or a date range."
                )

            with self.get_httpx_client() as client:
                http_response = client.get(
                    frankfurter_url,
                    params=params,
                )
                http_response.raise_for_status()
                result = http_response.json()
                return result, http_response
        except httpx.RequestError as e:
            raise ValueError(
                f"Failed to fetch historical exchange rates from {self.frankfurter_api_url}. {e}"
            )

    async def get_latest_exchange_rates(
        self,
        ctx: Context,
        base_currency: Annotated[
            ISO4217,
            Field(
                description="A base currency ISO4217 code for which rates are to be requested."
            ),
        ],
        symbols: Annotated[
            Optional[List[ISO4217] | ISO4217],
            Field(
                description="A list of target currency ISO4217 codes for which rates against the base currency will be provided. If not provided, all supported currencies will be shown."
            ),
        ] = None,
    ):
        """
        Returns the latest exchange rates for specific currencies. The
        symbols can be used to filter the results to specific currencies.
        If symbols is not provided, all supported currencies will be returned.
        """
        # Some LLMs make this mistake of passing just one currency but not as a list!
        if type(symbols) is str:
            symbols = [symbols]
        await ctx.info(
            f"Fetching latest exchange rates from Frankfurter API at {self.frankfurter_api_url}"
        )
        result, http_response = self._get_latest_exchange_rates(
            base_currency=base_currency,
            symbols=symbols,
        )
        return self.get_response_text_content(
            response=result, http_response=http_response
        )

    async def convert_currency_latest(
        self,
        ctx: Context,
        amount: Annotated[
            PositiveFloat,
            Field(description="The amount in the source currency to convert."),
        ],
        from_currency: Annotated[
            ISO4217, Field(description="The source currency ISO4217 code.")
        ],
        to_currency: Annotated[
            ISO4217, Field(description="The target currency ISO4217 code.")
        ],
    ):
        """
        Converts an amount from one currency to another using the latest exchange rates.
        """
        if from_currency.lower() == to_currency.lower():
            # If the source and target currencies are the same, no conversion is needed
            raise ValueError(
                f"Source currency '{from_currency}' and target currency '{to_currency}' are the same. No conversion needed."
            )
        await ctx.info(
            f"Obtaining latest exchange rates for {from_currency} to {to_currency} from Frankfurter API at {self.frankfurter_api_url}"
        )
        latest_rates, http_response = self._get_latest_exchange_rates(
            base_currency=from_currency,
            symbols=[to_currency],
        )
        await ctx.info(f"Converting {amount} of {from_currency} to {to_currency}")
        if not latest_rates or "rates" not in latest_rates:
            raise ValueError(
                f"Could not retrieve exchange rates for {from_currency} to {to_currency}."
            )
        rate = latest_rates["rates"].get(to_currency)
        if rate is None:
            raise ValueError(
                f"Exchange rate for {from_currency} to {to_currency} not found."
            )
        converted_amount = amount * float(rate)
        result = CurrencyConversionResponse(
            from_currency=from_currency,
            to_currency=to_currency,
            amount=amount,
            converted_amount=converted_amount,
            exchange_rate=rate,
            rate_date=latest_rates["date"],
        )
        return self.get_response_text_content(
            response=result, http_response=http_response
        )

    async def get_historical_exchange_rates(
        self,
        ctx: Context,
        base_currency: Annotated[
            ISO4217,
            Field(
                description="A base currency ISO4217 code for which rates are to be requested."
            ),
        ],
        symbols: Annotated[
            Optional[List[ISO4217] | ISO4217],
            Field(
                description="A list of target currency ISO4217 codes for which rates against the base currency will be provided. If not provided, all supported currencies will be shown."
            ),
        ] = None,
        specific_date: Annotated[
            Optional[date],
            Field(
                default=None,
                description="The specific date for which the historical rates are requested in the YYYY-MM-DD format.",
            ),
        ] = None,
        start_date: Annotated[
            Optional[date],
            Field(
                default=None,
                description="The start date, of a date range, for which the historical rates are requested in the YYYY-MM-DD format.",
            ),
        ] = None,
        end_date: Annotated[
            Optional[date],
            Field(
                default=None,
                description="The end date, of a date range, for which the historical rates are requested in the YYYY-MM-DD format.",
            ),
        ] = None,
    ):
        """
        Returns historical exchange rates for a specific date or date range.
        If the exchange rates for a specified date is not available, the rates available for
        the closest date before the specified date will be provided.
        Either a specific date, a start date, or a date range must be provided.
        The symbols can be used to filter the results to specific currencies.
        If symbols are not provided, all supported currencies will be returned.
        """
        await ctx.info(
            f"Fetching historical exchange rates from Frankfurter API at {self.frankfurter_api_url}"
        )
        # Some LLMs make this mistake of passing just one currency but not as a list!
        if type(symbols) is str:
            symbols = [symbols]
        result, http_response = self._get_historical_exchange_rates(
            specific_date=specific_date.isoformat() if specific_date else None,
            start_date=start_date.isoformat() if start_date else None,
            end_date=end_date.isoformat() if end_date else None,
            base_currency=base_currency,
            symbols=symbols,
        )
        await ctx.info(
            f"Historical exchange rates fetched for {len(result.get('rates', []))} dates."
        )
        return self.get_response_text_content(
            response=result, http_response=http_response
        )

    async def convert_currency_specific_date(
        self,
        ctx: Context,
        amount: Annotated[
            PositiveFloat,
            Field(description="The amount in the source currency to convert."),
        ],
        from_currency: Annotated[
            ISO4217, Field(description="The source currency ISO4217 code.")
        ],
        to_currency: Annotated[
            ISO4217, Field(description="The target currency ISO4217 code.")
        ],
        specific_date: Annotated[
            date,
            Field(
                description="The specific date for which the conversion is requested in the YYYY-MM-DD format."
            ),
        ],
    ):
        """
        Convert an amount from one currency to another using the exchange rates for a specific date.
        If there is no exchange rate available for the specific date, the rate for the closest available date before
        the specified date will be used.
        """
        if from_currency.lower() == to_currency.lower():
            # If the source and target currencies are the same, no conversion is needed
            raise ValueError(
                f"Source currency '{from_currency}' and target currency '{to_currency}' are the same. No conversion needed."
            )
        await ctx.info(
            f"Obtaining historical exchange rates for {from_currency} to {to_currency} on {specific_date} from Frankfurter API at {self.frankfurter_api_url}"
        )
        date_specific_rates, http_response = self._get_historical_exchange_rates(
            specific_date=specific_date.isoformat(),
            base_currency=from_currency,
            symbols=[to_currency],
        )
        await ctx.info(
            f"Converting {amount} of {from_currency} to {to_currency} on {specific_date}"
        )
        if not date_specific_rates or "rates" not in date_specific_rates:
            raise ValueError(
                f"Could not retrieve exchange rates for {from_currency} to {to_currency} for {specific_date}."
            )
        rate = date_specific_rates["rates"].get(to_currency)
        if rate is None:
            raise ValueError(
                f"Exchange rate for {from_currency} to {to_currency} not found."
            )
        converted_amount = amount * float(rate)
        result = CurrencyConversionResponse(
            from_currency=from_currency,
            to_currency=to_currency,
            amount=amount,
            converted_amount=converted_amount,
            exchange_rate=rate,
            rate_date=date_specific_rates["date"],
        )
        return self.get_response_text_content(
            response=result, http_response=http_response
        )


def app() -> FastMCP:
    app = FastMCP(
        name=AppMetadata.package_metadata["Name"],
        instructions=AppMetadata.package_metadata["Description"],
        on_duplicate_prompts="error",
        on_duplicate_resources="error",
        on_duplicate_tools="error",
    )
    mcp_obj = FrankfurterMCP()
    app_with_features = mcp_obj.register_features(app)
    return app_with_features


def main():  # pragma: no cover
    try:
        mcp_app = app()
        transport_type = env.str(
            name=EnvVar.MCP_SERVER_TRANSPORT,
            default=EnvVar.DEFAULT__MCP_SERVER_TRANSPORT,
            validate=OneOf(EnvVar.ALLOWED__MCP_SERVER_TRANSPORT),
        )
        (
            mcp_app.run(
                transport=transport_type,
                host=env.str(
                    name=EnvVar.FASTMCP_HOST,
                    default=EnvVar.DEFAULT__FASTMCP_HOST,
                ),
                port=env.int(
                    name=EnvVar.FASTMCP_PORT,
                    default=EnvVar.DEFAULT__FASTMCP_PORT,
                ),
                uvicorn_config={
                    "timeout_graceful_shutdown": 5,  # seconds
                },
            )
            if transport_type != "stdio"
            else mcp_app.run(transport=transport_type)
        )
    except KeyboardInterrupt:
        sys.exit(0)
    except Exception as e:
        logger.error(e)
        sys.exit(1)
    finally:
        # Cleanup if necessary
        pass


if __name__ == "__main__":
    main()



from frankfurtermcp.common import EnvVar, AppMetadata
from frankfurtermcp.model import CurrencyConversionResponse, ResponseMetadata


class TestDataModels:
    def test_currency_conversion_response(self):
        """
        Test the CurrencyConversionResponse data model to ensure that it correctly validates and stores data.
        """
        # Made up data for testing
        data = {
            "from_currency": "GBP",
            "to_currency": "EUR",
            "amount": 100.0,
            "converted_amount": 115.0,
            "exchange_rate": 1.15,
            "rate_date": "2025-08-31",
        }
        response = CurrencyConversionResponse(**data)
        assert response.from_currency == "GBP"
        assert response.to_currency == "EUR"
        assert response.amount == 100.0
        assert response.converted_amount == 115.0
        assert response.exchange_rate == 1.15
        assert response.rate_date.isoformat() == "2025-08-31"

    def test_currency_conversion_response_negative_floats(self):
        """
        Test the CurrencyConversionResponse data model to ensure that it raises a validation error for invalid amount.
        """
        data = {
            "from_currency": "GBP",
            "to_currency": "EUR",
            "amount": -100.0,  # Invalid negative amount
            "converted_amount": -115.0,  # Invalid negative amount
            "exchange_rate": -1.15,  # Invalid negative exchange rate
            "rate_date": "2025-08-31",
        }
        obj: CurrencyConversionResponse = None
        try:
            obj = CurrencyConversionResponse(**data)
        except Exception as e:
            assert (
                f"3 validation errors for {CurrencyConversionResponse.__name__}"
                in str(e)
            )
        finally:
            assert obj is None

    def test_currency_conversion_response_invalid_currency_codes(self):
        """
        Test the CurrencyConversionResponse data model to ensure that it raises a validation error for invalid currency codes.
        """
        data = {
            "from_currency": "ABC",  # Invalid currency code
            "to_currency": "XYZ",
            "amount": 100.0,
            "converted_amount": 115.0,
            "exchange_rate": 1.15,
            "rate_date": "2025-08-31",
        }
        obj: CurrencyConversionResponse = None
        try:
            obj = CurrencyConversionResponse(**data)
        except Exception as e:
            assert (
                f"2 validation errors for {CurrencyConversionResponse.__name__}"
                in str(e)
            )
        finally:
            assert obj is None

    def test_currency_conversion_response_invalid_date(self):
        """
        Test the CurrencyConversionResponse data model to ensure that it raises a validation error for invalid date format.
        """
        data = {
            "from_currency": "GBP",
            "to_currency": "EUR",
            "amount": 100.0,
            "converted_amount": 115.0,
            "exchange_rate": 1.15,
            "rate_date": "25-32-31",  # Invalid date format
        }
        obj: CurrencyConversionResponse = None
        try:
            obj = CurrencyConversionResponse(**data)
        except Exception as e:
            assert (
                f"1 validation error for {CurrencyConversionResponse.__name__}"
                in str(e)
            )
        finally:
            assert obj is None

    def test_response_metadata(self):
        """
        Test the ResponseMetadata data model to ensure that it correctly initializes with valid data.
        """

        data = {
            "version": AppMetadata.package_metadata["Version"],
            "api_url": EnvVar.DEFAULT__FRANKFURTER_API_URL,
            "api_status_code": 200,
            "api_bytes_downloaded": 512,
            "api_elapsed_time": 150000,  # in microseconds
        }
        response = ResponseMetadata(**data)
        assert response.version == AppMetadata.package_metadata["Version"]
        assert str(response.api_url) == EnvVar.DEFAULT__FRANKFURTER_API_URL
        assert response.api_status_code == 200
        assert response.api_bytes_downloaded == 512
        assert response.api_elapsed_time == 150000

    def test_response_metadata_negative_values(self):
        """
        Test the ResponseMetadata data model to ensure that it raises a validation error for negative values.
        """

        data = {
            "version": AppMetadata.package_metadata["Version"],
            "api_url": EnvVar.DEFAULT__FRANKFURTER_API_URL,
            "api_status_code": 200,
            "api_bytes_downloaded": -512,  # Invalid negative bytes downloaded
            "api_elapsed_time": -150000,  # Invalid negative elapsed time
        }
        obj: ResponseMetadata = None
        try:
            obj = ResponseMetadata(**data)
        except Exception as e:
            assert f"2 validation errors for {ResponseMetadata.__name__}" in str(e)
        finally:
            assert obj is None

    def test_response_metadata_invalid_url(self):
        """
        Test the ResponseMetadata data model to ensure that it raises a validation error for invalid URL.
        """

        data = {
            "version": AppMetadata.package_metadata["Version"],
            "api_url": "htp:/invalid-url",  # Invalid URL format
            "api_status_code": 200,
            "api_bytes_downloaded": 512,
            "api_elapsed_time": 150000,  # in microseconds
        }
        obj: ResponseMetadata = None
        try:
            obj = ResponseMetadata(**data)
        except Exception as e:
            assert f"1 validation error for {ResponseMetadata.__name__}" in str(e)
        finally:
            assert obj is None



from typing import Annotated
from pydantic import BaseModel, Field, HttpUrl, PositiveFloat, PositiveInt
from pydantic_extra_types.currency_code import ISO4217
from datetime import date


class CurrencyConversionResponse(BaseModel):
    """
    Response model for currency conversion.
    """

    from_currency: ISO4217 = Field(
        description="The ISO 4217 code of the currency to convert from."
    )
    to_currency: ISO4217 = Field(
        description="The ISO 4217 code of the currency to convert to."
    )
    amount: PositiveFloat = Field(
        description="The amount (of the source currency) to convert."
    )
    converted_amount: PositiveFloat = Field(
        description="The converted amount (of the target currency)."
    )
    exchange_rate: PositiveFloat = Field(
        description="The exchange rate used for the conversion."
    )
    rate_date: date = Field(
        description="The date, in ISO format, of the exchange rate used for the conversion."
    )


class ResponseMetadata(BaseModel):
    """
    Metadata for MCP response.
    """

    version: Annotated[str, Field(description="The version of the package.")]
    api_url: HttpUrl = (
        Field(description="The URL of the Frankfurter API used for the call."),
    )
    api_status_code: int = Field(
        description="The HTTP status code of the API response."
    )
    api_bytes_downloaded: PositiveInt = Field(
        description="The number of bytes downloaded in the API response."
    )
    api_elapsed_time: PositiveInt = Field(
        description="The elapsed time for the API call in microseconds."
    )



import json
import logging
import os
import ssl
from typing import Any, ClassVar, Dict, List

import certifi
from fastmcp import FastMCP
import copy

import httpx
from mcp.types import TextContent
from pydantic import BaseModel

from frankfurtermcp.common import AppMetadata, EnvVar
from frankfurtermcp.model import ResponseMetadata

from frankfurtermcp import env

logger = logging.getLogger(__name__)


class MCPMixin:
    """
    A mixin class to register tools, resources, and prompts with a FastMCP instance.
    """

    # Each entry is a dict, must include "fn" (method name),
    # rest is arbitrary metadata relevant to FastMCP.
    tools: ClassVar[List[Dict[str, Any]]] = []
    # Each entry is a dict, must include "fn" (method name) and "uri",
    # rest is arbitrary metadata relevant to FastMCP.
    resources: ClassVar[List[Dict[str, Any]]] = []
    # Each entry is a dict, must include "fn" (method name),
    # rest is arbitrary metadata relevant to FastMCP.
    prompts: ClassVar[List[Dict[str, Any]]] = []

    frankfurter_api_url: ClassVar[str] = env.str(
        name=EnvVar.FRANKFURTER_API_URL,
        default=EnvVar.DEFAULT__FRANKFURTER_API_URL,
    )

    def register_features(self, mcp: FastMCP) -> FastMCP:
        """
        Register tools, resources, and prompts with the given FastMCP instance.

        Args:
            mcp (FastMCP): The FastMCP instance to register features with.

        Returns:
            FastMCP: The FastMCP instance with registered features.
        """
        # Register tools
        for tool in self.tools:
            assert "fn" in tool, "Tool metadata must include the 'fn' key."
            tool_copy = copy.deepcopy(tool)
            fn_name = tool_copy.pop("fn")
            fn = getattr(self, fn_name)
            mcp.tool(**tool_copy)(fn)  # pass remaining metadata as kwargs
            logger.debug(f"Registered MCP tool: {fn_name}")
        # Register resources
        for res in self.resources:  # pragma: no cover
            assert "fn" in res and "uri" in res, (
                "Resource metadata must include 'fn' and 'uri' keys."
            )
            res_copy = copy.deepcopy(res)
            fn_name = res_copy.pop("fn")
            uri = res_copy.pop("uri")
            fn = getattr(self, fn_name)
            mcp.resource(uri, **res_copy)(fn)
            logger.debug(f"Registered MCP resource at URI: {uri}")
        # Register prompts
        for pr in self.prompts:  # pragma: no cover
            assert "fn" in pr, "Prompt metadata must include the 'fn' key."
            pr_copy = copy.deepcopy(pr)
            fn_name = pr_copy.pop("fn")
            fn = getattr(self, fn_name)
            mcp.prompt(**pr_copy)(fn)
            logger.debug(f"Registered MCP prompt: {fn_name}")

        return mcp

    def get_response_text_content(
        self,
        response: Any,
        http_response: httpx.Response,
        include_metadata: bool = env.bool(
            name=EnvVar.MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE,
            default=EnvVar.DEFAULT__MCP_SERVER_INCLUDE_METADATA_IN_RESPONSE,
        ),
    ) -> TextContent:
        """
        Convert response data to TextContent format.

        Args:
            response (Any): The response data to convert.
            http_response (httpx.Response): The HTTP response object for header extraction.
            include_metadata (bool): Whether to include metadata in the TextContent.

        Returns:
            TextContent: The converted TextContent object.
        """
        literal_text = "text"
        if isinstance(response, TextContent):
            # do nothing yet
            pass
        elif isinstance(response, (str, int, float, complex, bool, type(None))):
            text_content = TextContent(type=literal_text, text=str(response))
        elif isinstance(response, dict) or isinstance(response, list):
            text_content = TextContent(type=literal_text, text=json.dumps(response))
        elif isinstance(response, BaseModel):
            text_content = TextContent(
                type=literal_text, text=response.model_dump_json()
            )
        else:
            raise TypeError(
                f"Unsupported data type: {type(response).__name__}. "
                "Only str, int, float, complex, bool, dict, list, and Pydantic BaseModel types are supported for wrapping as TextContent."
            )
        if include_metadata:
            text_content.meta = (
                text_content.meta if hasattr(text_content, "_meta") else {}
            )
            text_content.meta[AppMetadata.PACKAGE_NAME] = ResponseMetadata(
                version=AppMetadata.package_metadata["Version"],
                api_url=self.frankfurter_api_url,
                api_status_code=http_response.status_code,
                api_bytes_downloaded=http_response.num_bytes_downloaded,
                api_elapsed_time=http_response.elapsed.microseconds,
            ).model_dump()
        return text_content


class HTTPHelperMixin:
    """
    A mixin class to provide HTTP client functionality using httpx.
    """

    def get_httpx_client(self) -> httpx.Client:
        """
        Obtain an HTTPX client for making requests.
        """
        verify = env.bool(
            name=EnvVar.HTTPX_VERIFY_SSL,
            default=EnvVar.DEFAULT__HTTPX_VERIFY_SSL,
        )
        if verify is False:
            logging.warning(
                "SSL verification is disabled. This is not recommended for production use."
            )
        ctx = ssl.create_default_context(
            cafile=os.environ.get("SSL_CERT_FILE", certifi.where()),
            capath=os.environ.get("SSL_CERT_DIR"),
        )
        client = httpx.Client(
            verify=verify if (verify is not None and verify is False) else ctx,
            follow_redirects=True,
            trust_env=True,
            timeout=env.float(
                name=EnvVar.HTTPX_TIMEOUT,
                default=EnvVar.DEFAULT__HTTPX_TIMEOUT,
            ),
        )
        return client
